# Load packages
# The code for nodewise Lasso was written by Ruben Dezeure

library(glmnet)
library(scalreg)
library(SIS)
source(file.path("Nodewise-Lasso.R"))

# This function implements the method for testing sparse signals in Zhang and Cheng (2016)
# Input: n times p design matrix X.f, response variable Y.f, the (sub)sample size used for screening, the set of variables of interest, 
# the number of bootstrap replications (default 500), the nominal level alpha (default 0.05)
# Output: values of the non-studentized and studentized statistics, and whether the tests get rejected at the level alpha

ST <- function(X.f, Y.f, test.set, M=500, alpha=0.05)
{
n <- dim(X.f)[1]
p <- dim(X.f)[2]

#n1 <- sub.size
#n0 <- n-floor(n1)
#S1 <- sample(1:n, floor(n1), replace=FALSE)
#X.sub <- X.f[S1,]
#Y.sub <- Y.f[S1]
#cvfit <- cv.glmnet(X.sub, Y.sub, intercept=FALSE)
#cf <- as.numeric(coef(cvfit, s="lambda.min"))[-1]
#set1 <- (1:p)[abs(cf)>0]
#resi <- Y.sub-X.sub%*%cf
#beta.m <- t(standardize(X.sub[,-set1]))%*%resi
#screen.set <- sort(order(abs(beta.m),decreasing=TRUE)[1:(n0-1-length(set1))])
#a <- (1:p)[-set1]
#screen.set <- union(a[screen.set],set1)
X <- X.f#[-S1,screen.set]
Y <- Y.f#[-S1]

node <- score.nodewiselasso(X, wantTheta=TRUE, verbose=FALSE, lambdaseq="quantile",
parallel=FALSE, ncores=2, oldschool = FALSE, lambdatuningfactor = 1)
Theta <- node$out
Gram<-t(X)%*%X/n

sreg <- scalreg(X,Y)
beta.hat <- sreg$coefficients
sigma.sq <- sum((Y-X%*%beta.hat)^2)/(n-sum(abs(beta.hat)>0))
#test.set.i <- intersect(screen.set,test.set)
index <- test.set#screen.set%in%test.set.i

Omega <- diag(Theta%*%Gram%*%t(Theta))*sigma.sq
beta.db <- beta.hat+Theta%*%t(X)%*%(Y-X%*%beta.hat)/n
#margin.st <- sqrt(n)*abs(beta.db[index])/sqrt(Omega[index])
margin.nst <- sqrt(n)*abs(beta.db[index])
#stat.st <- max(margin.st)
stat.nst <- max(margin.nst)

stat.boot.st <- stat.boot.nst <- rep(NA,M)
for(i in 1:M)
 {
 e <- rnorm(n)
 xi.boot <- Theta[index,]%*%t(X)%*%e*sqrt(sigma.sq)/sqrt(n)
 stat.boot.nst[i] <- max(abs(xi.boot))
#stat.boot.st[i] <- max(abs(xi.boot/sqrt(Omega[index])))
 }
rej.nst <- rep(0, 2)
if(stat.nst>quantile(stat.boot.nst, 0.99)) rej.nst[1] <- 1 else rej.nst[1] <- 0
if(stat.nst>quantile(stat.boot.nst, 0.95)) rej.nst[2] <- 1 else rej.nst[2] <- 0

#result <- list(stat.nst, rej.nst, stat.st, rej.st)
#names(result) <- c("non-studentized test","non-studentized test","studentized test","studentized test")
return(rej.nst)
}

# An example

Gen.data <- function(n=500, p=50, tau=1, delta=Inf){
  X <- matrix(rnorm(n*p), n, p)
  A <- rbinom(n, 1, 0.5)
  # generate the baseline
  h0 <- 1+(X[,1]-X[,2])/2
  # generate the contrast function
  beta0 <- rep(1, p)/sqrt(p)
  if (tau==1){
    tau0 <- delta*sqrt(pi)*(X[,1]+X[,2])
  }
  else if (tau==2){
    tau0 <- 0.5*((X[,1]+X[,2])^2-delta)
#	tau0 <- 0.25*(X[,1]+X[,2])^2
  }
  else if (tau==3){
	tau0 <- sqrt(pi)*(X[,1]+X[,2])^3/16
  }
  else{
    tau0 <- delta*((rowSums(X[,1:2]))^2/10-(rowSums(X[,3:20]))^2/18)*(rowSums(X[,21:25]))^2/5
  }
#  tau0 <- tau0*(X[,3]+X[,4]+X[,5]+X[,6]+X[,7])^2/5
  
  # generate the response
  Y <- h0+A*tau0+0.5*rnorm(n)
  Z <- X%*%beta0
  
  return(list(Y=Y, X=X, A=A, Z=Z))
}


set.seed(12345)
n <- 500
p <- 50
re <- matrix(0, 500, 2)
for (i in 1:500){
set.seed(12345*i)
Model <- Gen.data(n=1000, p=p, tau=4, delta=0.175*pi)
Y <- Model$Y
Y <- Model$Y; Y <- Y-mean(Y)
X <- cbind(Model$X, (Model$A)*cbind(1, Model$X)); X <- t(t(X)-colMeans(X))
test.set <- (p+2):(2*p+1)
re[i,] <- ST(X, Y, test.set)
}
save.image(file="DLS16p50.RData")

