## simulation code for testing high dimensional qualitative treatment effects
## Chengchun Shi, April 17, 2017

library(ncvreg)
dyn.load("RPtest.so")
## data generation function
## tau: the specification of contrast function
## if tau=1, tau=2Xbeta I(Xbeta>0)+2min(Xbeta+delta,0)
## else tau=(Xbeta)^2-delta
## return the main subspace
Gen.data <- function(n=500, p=50, tau=1, delta=Inf){
  X <- matrix(0, n, p)
  for (i in 1:n){
      X[i, ] <- arima.sim(model=list(ar=0.5), sd=sqrt(0.75), n=p)}
  A <- rbinom(n, 1, 0.5)
  # generate the baseline
  h0 <- 1+(X[,1]-X[,2])/2
  # generate the contrast function
  beta0 <- rep(1, p)/sqrt(p)
  if (tau==1){
    tau0 <- delta*sqrt(pi)*(X[,1]+X[,2])
  }
  else if (tau==2){
#    tau0 <- 0.5*((X[,1]+X[,2])^2-delta)
	tau0 <- 0.25*(X[,1]+X[,2])^2
  }
  else if (tau==3){
	tau0 <- sqrt(pi)*(X[,1]+X[,2])^3/16
  }
  else{
	tau0 <- delta*((rowSums(X[,1:2]))^2/2-(rowSums(X[,3:20]))^2/18)*(rowSums(X[,21:25]))^2/5
  }
#  tau0 <- tau0*(X[,3]+X[,4]+X[,5]+X[,6]+X[,7])^2/5
  
  # generate the response
  Y <- h0+A*tau0+0.5*rnorm(n)
  Z <- X%*%beta0
  
  return(list(Y=Y, X=X, A=A, Z=Z))
}

## estimate the propensity score and baseline function
## using sparse penalized regression with SCAD penalty
pih <- function(X, A, Y){
  
  p <- dim(X)[2]
  
  # estimate the baseline
  theta <- cv.ncvreg(X[A==0,], Y[A==0], penalty='SCAD', family="gaussian", nfolds=5)
  theta.est <- coef(theta, s='lambda.min')
  theta.est <- theta.est[1:(p+1)]
  h0.est <- cbind(1,X)%*%theta.est
  
  # estimate the baseline
  zeta <- cv.ncvreg(X[A==1,], Y[A==1], penalty='SCAD', family="gaussian", nfolds=5)
  zeta.est <- coef(zeta, s='lambda.min')
  zeta.est <- zeta.est[1:(p+1)]
  h1.est <- cbind(1,X)%*%zeta.est
  
  # estimate the propensity score
  alpha <- cv.ncvreg(X, A, family="binomial", penalty='SCAD', nfolds=5)
  alpha.est <- coef(alpha, s='lambda.min')
  pi.est <- plogis(as.vector(cbind(1,X) %*% as.matrix(alpha.est)))
  
  M.alpha <- alpha.est!=0
  M.theta0 <- theta.est!=0
  M.theta1 <- zeta.est!=0
  
  return(list(pi.est=pi.est, h0.est=h0.est, h1.est=h1.est, M.alpha=M.alpha, 
              M.theta0=M.theta0, M.theta1=M.theta1, alpha.est=alpha.est,
              theta.est=theta.est, zeta.est=zeta.est))
}

## compute the pseodu outcome
tau <- function(X, A, Y, est.pih){
  
  pi.est <- est.pih$pi.est
  h0.est <- est.pih$h0.est
  h1.est <- est.pih$h1.est
  tau.est <- (A/pi.est-(1-A)/(1-pi.est))*Y
  tau.est <- tau.est-(A/pi.est-1)*h1.est+((1-A)/(1-pi.est)-1)*h0.est
  
  return(tau.est)
}

# ## compute the value difference
# VD_r <- function(X, A, Y, est.pih, tau.est, I){
#   X <- X[I,]
#   A <- A[I]
#   Y <- Y[I]
#   
#   n <- length(A)
#   pi.est <- est.pih$pi.est[I]
#   h0.est <- est.pih$h0.est[I]
#   h1.est <- est.pih$h1.est[I]
#   
#   ## compute the estimator
#   d.est <- (tau.est>0)
#   VD.est <- ((1-A)/(1-pi.est)-A/(pi.est))*Y
#   VD.est <- VD.est-((1-A)/(1-pi.est)-1)*h0.est+(A/pi.est-1)*h1.est
#   kappa <- VD.est
#   VD.est <- VD.est*(1-d.est)
#   
#   return(mean(VD.est))
# }

## compute the test statistic
stat <- function(X, A, Y, est.pih, dn, tau.est){
  n <- length(A)
  pi.est <- est.pih$pi.est
  h0.est <- est.pih$h0.est
  h1.est <- est.pih$h1.est
  M.alpha <- est.pih$M.alpha
  M.theta0 <- est.pih$M.theta0
  M.theta1 <- est.pih$M.theta1
  
  ## compute the numerator
  d.est <- (tau.est>0)
  if (dim(tau.est)[2]>1)
    d.est <- rowMeans(d.est)
  else
    d.est <- as.vector(d.est)
  VD.est <- ((1-A)/(1-pi.est)-A/(pi.est))*Y
  VD.est <- VD.est-((1-A)/(1-pi.est)-1)*h0.est+(A/pi.est-1)*h1.est
  #  kappa <- VD.est
  VD.est <- VD.est*(1-d.est)
  kappa <- VD.est
  
  ## compute the denominator
  Sigma2 <- t(cbind(1,X)[,M.theta0]) %*% ((1-A)*cbind(1,X)[,M.theta0])/n
  Sigma3 <- t(cbind(1,X)[,M.theta1]) %*% (A*cbind(1,X)[,M.theta1])/n
  if (length(which(M.theta0))==1)
    barI2 <- mean(as.vector(((1-A)/(1-pi.est)-1)) * (cbind(1,X)[,M.theta0]) * (1-d.est))
  else
    barI2 <- colMeans(as.vector(((1-A)/(1-pi.est)-1)) * (cbind(1,X)[,M.theta0]) * (1-d.est))
  if (length(which(M.theta1))==1)
    barI3 <- mean(as.vector(A/pi.est-1) * (cbind(1,X)[,M.theta1]) * (1-d.est))
  else
    barI3 <- colMeans(as.vector(A/pi.est-1) * (cbind(1,X)[,M.theta1]) * (1-d.est))
  SiginvbarI2 <- solve(Sigma2, barI2)
  SiginvbarI3 <- solve(Sigma3, barI3)
  
  if (length(SiginvbarI2)>1)
    kappa <- kappa-(as.vector((1-A)*(Y-h0.est))*(cbind(1,X)[,M.theta0]))%*%SiginvbarI2
  else
    kappa <- kappa-(as.vector((1-A)*(Y-h0.est))*(cbind(1,X)[,M.theta0]))*SiginvbarI2
  if (length(SiginvbarI3)>1)
    kappa <- kappa+(as.vector(A*(Y-h1.est))*(cbind(1,X)[,M.theta1]))%*%SiginvbarI3
  else
    kappa <- kappa+(as.vector(A*(Y-h1.est))*(cbind(1,X)[,M.theta1]))*SiginvbarI3
  if (length(which(M.alpha))>0){
    Sigma1 <- t(cbind(1,X)[,M.alpha]) %*% (pi.est*(1-pi.est)*cbind(1,X)[,M.alpha])/n
    if (length(which(M.alpha))==1){
      barI1 <- mean(as.vector((pi.est*(1-A)/(1-pi.est))*(Y-h0.est)+((1-pi.est)*A/pi.est)*(Y-h1.est))
                    *cbind(1,X)[,M.alpha]*(1-d.est))
    }
    else{
      barI1 <- colMeans(as.vector((pi.est*(1-A)/(1-pi.est))*(Y-h0.est)+((1-pi.est)*A/pi.est)*(Y-h1.est))
                        *cbind(1,X)[,M.alpha]*(1-d.est))
    }
    SiginvbarI1 <- solve(Sigma1, barI1)
    if (length(SiginvbarI1)>1)
      kappa <- kappa+(A-pi.est)*cbind(1,X)[,M.alpha]%*%SiginvbarI1
    else
      kappa <- kappa+(A-pi.est)*cbind(1,X)[,M.alpha]*SiginvbarI1
  }
  #  kappa <- kappa*(1-d.est)
  sig.est <- max(sqrt(var(kappa)), dn)
  
  return(sqrt(n)*mean(VD.est)/sig.est)
}

## Random projection ensemble test
# B the number of random projections generated to take supremum
# m the number of random projections generated to take average
# K the number of cross validation folds
test.RPCV <- function(Model, dn, B=100, m=200, K=5, iter=1){
  N <- length(Model$A)
  n <- N/2
  p <- dim(Model$X)[2]
  
  ## partition the data into two subsets
  I1 <- sample(1:N, n)
  I2 <- setdiff(1:N, I1)
  
  ## estimate the propensity score and baseline for each subset
  est.pih1 <- pih(Model$X[I1,], Model$A[I1], Model$Y[I1])
  est.pih2 <- pih(Model$X[I2,], Model$A[I2], Model$Y[I2])
  
  ## construct the pseodu outcome
  est.tau <- rep(0, n)
  est.tau[I1] <- tau(Model$X[I1,], Model$A[I1], Model$Y[I1], est.pih1)
  est.tau[I2] <- tau(Model$X[I2,], Model$A[I2], Model$Y[I2], est.pih2)
  
  ## cross validation
  folds <- split(sample(n, n, replace=FALSE), as.factor(1:K))
  ITrain <- NULL
  ITest <- NULL
  for (i in 1:K){
    ITrain <- c(ITrain, setdiff(1:n, folds[[i]]))
    ITest <- c(ITest, folds[[i]])
  }
  ITrain <- ITrain-1
  ITest <- ITest-1
  
  ## predict the response
  result <- .C("GS", as.double(rep(0, n*m)), as.double(t(Model$X[I2,])), as.double(t(Model$X[I1,])), 
               as.double(Model$A[I1]), as.double(Model$Y[I1]), as.double(c(est.pih1$pi.est, 
                                                                           est.pih1$h0.est, est.pih1$h1.est)), as.double(est.tau[I1]), as.integer(ITrain), 
               as.integer(ITest), as.integer(c(n, K, B, m, p, 3, n/K)), as.integer(2*iter-2))
  tau.est.all <- matrix(result[[1]], n, m)
  
  # construct the test statistic
  stat1 <- stat(Model$X[I2,], Model$A[I2], Model$Y[I2], est.pih2, dn, tau.est.all)
  
  ## cross validation
  folds <- split(sample(n, n, replace=FALSE), as.factor(1:K))
  ITrain <- NULL
  ITest <- NULL
  for (i in 1:K){
    ITrain <- c(ITrain, setdiff(1:n, folds[[i]]))
    ITest <- c(ITest, folds[[i]])
  }
  ITrain <- ITrain-1
  ITest <- ITest-1
  
  ## predict the response
  result <- .C("GS", as.double(rep(0, n*m)), as.double(t(Model$X[I1,])), as.double(t(Model$X[I2,])), 
               as.double(Model$A[I2]), as.double(Model$Y[I2]), as.double(c(est.pih2$pi.est, 
                                                                           est.pih2$h0.est, est.pih2$h1.est)), as.double(est.tau[I2]), as.integer(ITrain), 
               as.integer(ITest), as.integer(c(n, K, B, m, p, 3, n/K)), as.integer(2*iter-1))
  tau.est.all <- matrix(result[[1]], n, m)
  
  # construct the test statistic
  stat2 <- stat(Model$X[I1,], Model$A[I1], Model$Y[I1], est.pih1, dn, tau.est.all)
  
  # take the supremum of two test statistics
  stat0 <- max(stat1, stat2)
  return(stat0)
}

# GO, PACK!
stat0 <- rep(0, 600)
dn <- log(log(1000, 10))/(1000^(1/6))
for (l in 251:300) {
  #do your work
  print(l)
  
  #sleep to make this take extra time
  set.seed(12345*l)
  
  Model <- Gen.data(n=1000, p=100, tau=4, delta=0.1510546)
  stat0[l] <- test.RPCV(Model, dn, B=400000, m=1, iter=l)
  
  #save the data every 50 iterations
  if (l %% 50 == 0) {
    #save all of the data
    save.image(file="RPCVC5S3p100part6.RData")
    
    print("saved")
  }	
}
