/* some C subroutines in simulation for testing high dimensional qualitative treatment effects: 
Inference for Gaussian process using approximated restricted likelihood */

// include libraries
#include<stdio.h>
#include<time.h>
#include<math.h>
#include<R.h>
#include<Rinternals.h>
#include<R_ext/Rdynload.h>
#include<gsl/gsl_vector.h>
#include<gsl/gsl_matrix.h>
#include<gsl/gsl_blas.h>
#include<gsl/gsl_linalg.h>
#include<gsl/gsl_errno.h>
#include<gsl/gsl_rng.h>
#include<gsl/gsl_randist.h>
#include <gsl/gsl_bspline.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_statistics.h>

extern "C"{

// return x*(1-(y>0))
double ghf(double x, double y){
	if (y>0)
		return 0.0;
	else
		return x;
}

// calculate the cross-validated value difference
// est_pih: m*3 matrix
// m: number of elements in I
// n: number of elements in A
void VD(double *VD_est, double *A, double *Y, double *est_pih, double *tau_est, int *I, int *ntest, int *n)
{
	try{
		// vector allocation 
		gsl_vector *gsl_vd_est = gsl_vector_alloc(*ntest);
		double *vd_est = gsl_vector_ptr(gsl_vd_est, 0);

		// calculate the value difference
		for (int i=0;i<*ntest;i++){
			vd_est[i] = ((1-A[I[i]])/(1-est_pih[I[i]])-A[I[i]]/est_pih[I[i]])*Y[I[i]];						//  VD=((1-A)/pi-A/pi)*Y 
			vd_est[i] = vd_est[i]-((1-A[I[i]])/(1-est_pih[I[i]])-1)*est_pih[(*n)+I[i]]
						+(A[I[i]]/est_pih[I[i]]-1)*est_pih[(*n)*2+I[i]];									//	VD=VD-((1-A)/(1-pi)-1)*h0+(A/pi-1)*h1
			vd_est[i] = ghf(vd_est[i], tau_est[I[i]]);
		}

		// sum the value differenece
		for (int i=0;i<*ntest;i++){
			VD_est[0] = VD_est[0]+vd_est[i];
		}

		// free the allocation
		gsl_vector_free(gsl_vd_est);
	}
	catch (...) {
		::Rf_error( "c++ exception (unknown reason)" );
	}
}

// Fit B-spline
// Z: projected covariates
// est.tau: pseduo response
// tau.est: the predicted response
// Itrain: the index of training samples
// Itest: the index of testing samples
// nnod: number of nods in spline
// m: number of testing samples
void BS(double *tau_est, double *Z, double *est_tau, int *Itrain, int *Itest, int *nnod, int *m, int *n)
{
	try{
		/*initialize*/
		int m0 = (*n)-(*m);
		gsl_bspline_workspace *gsl_bs = gsl_bspline_alloc(4, *nnod+2);

		/*use quantiles as the breakpoints*/
		gsl_vector *gsl_bp = gsl_vector_alloc(*nnod+2);
		gsl_vector *gsl_Z2 = gsl_vector_alloc(*n);
		gsl_vector_view gsl_Z=gsl_vector_view_array(Z, *n);
		gsl_vector_memcpy(gsl_Z2, &gsl_Z.vector);
		double *Z2_ptr = gsl_vector_ptr(gsl_Z2, 0);
		gsl_sort(Z2_ptr, 1, *n);

		gsl_vector_set(gsl_bp, 0, Z2_ptr[0]);
		for (int i=0;i<(*nnod);i++){
			gsl_vector_set(gsl_bp, i+1, gsl_stats_quantile_from_sorted_data(Z2_ptr, 1, *n, (1.0*(i+1))/(*nnod+1)));
		}
		gsl_vector_set(gsl_bp, *nnod+1, Z2_ptr[*n-1]);
		gsl_bspline_knots(gsl_bp, gsl_bs);

		/*construct the fit matrix X*/
		int ncoef = *nnod+4;
		gsl_vector *gsl_B = gsl_vector_alloc(ncoef);
		gsl_matrix *gsl_X = gsl_matrix_alloc(m0, ncoef);
		gsl_matrix *gsl_Cov = gsl_matrix_alloc(ncoef, ncoef);
		for (int i=0;i<m0;i++){
			/* compute B_j(xi) for all j */
			gsl_bspline_eval(Z[Itrain[i]], gsl_B, gsl_bs);
			/* fill in row i of X */
			for (int j=0;j<ncoef;j++){
				gsl_matrix_set(gsl_X, i, j, gsl_vector_get(gsl_B, j));
			}
		}

		/*construct the response*/
		gsl_vector *gsl_y = gsl_vector_alloc(m0);
		for (int i=0;i<m0;i++){
			gsl_vector_set(gsl_y, i, est_tau[Itrain[i]]);
		}

		/*least square fit*/
		gsl_vector *reg_coef = gsl_vector_alloc(ncoef);
		gsl_multifit_linear_workspace *mw = gsl_multifit_linear_alloc(m0, ncoef);
		double chisq;
		gsl_multifit_linear(gsl_X, gsl_y, reg_coef, gsl_Cov, &chisq, mw);

		/*prediction*/
		double yerr;
		for (int i=0;i<*m;i++){
			gsl_bspline_eval(Z[Itest[i]], gsl_B, gsl_bs);
			gsl_multifit_linear_est(gsl_B, reg_coef, gsl_Cov, &tau_est[Itest[i]], &yerr);
		}

		// free the allocation
		gsl_vector_free(gsl_B);
		gsl_vector_free(gsl_Z2);
		gsl_vector_free(reg_coef);
		gsl_vector_free(gsl_bp);
		gsl_matrix_free(gsl_X);
		gsl_matrix_free(gsl_Cov);
		gsl_bspline_free(gsl_bs);
		gsl_vector_free(gsl_y);
		gsl_multifit_linear_free(mw);
	}
	catch (...) {
		::Rf_error( "c++ exception (unknown reason)" );
	}
}

// Fit B-spline version II
void BSII(double *tau_est, double *Ztest, double *Z, double *est_tau, int *nnod, int *n)
{
	try{
		/*initialize*/
		gsl_bspline_workspace *gsl_bs = gsl_bspline_alloc(4, *nnod+2);

		/*use quantiles as the breakpoints*/
		gsl_vector *gsl_bp = gsl_vector_alloc(*nnod+2);
		gsl_vector *gsl_Z2 = gsl_vector_alloc((*n)*2);
		double *Z2_ptr = gsl_vector_ptr(gsl_Z2, 0);
		for (int i=0;i<*n;i++)
			Z2_ptr[i] = Z[i];
		for (int i=0;i<*n;i++)
			Z2_ptr[i+(*n)] = Ztest[i];
		gsl_sort(Z2_ptr, 1, 2*(*n));

		gsl_vector_set(gsl_bp, 0, Z2_ptr[0]);
		for (int i=0;i<(*nnod);i++){
			gsl_vector_set(gsl_bp, i+1, gsl_stats_quantile_from_sorted_data(Z2_ptr, 1, 2*(*n), (1.0*(i+1))/(*nnod+1)));
		}
		gsl_vector_set(gsl_bp, *nnod+1, Z2_ptr[2*(*n)-1]);
		gsl_bspline_knots(gsl_bp, gsl_bs);

		/*construct the fit matrix X*/
		int ncoef = *nnod+4;
		gsl_vector *gsl_B = gsl_vector_alloc(ncoef);
		gsl_matrix *gsl_X = gsl_matrix_alloc(*n, ncoef);
		gsl_matrix *gsl_Cov = gsl_matrix_alloc(ncoef, ncoef);
		for (int i=0;i<*n;i++){
			/* compute B_j(xi) for all j */
			gsl_bspline_eval(Z[i], gsl_B, gsl_bs);
			/* fill in row i of X */
			for (int j=0;j<ncoef;j++){
				gsl_matrix_set(gsl_X, i, j, gsl_vector_get(gsl_B, j));
			}
		}

		/*construct the response*/
		gsl_vector_view gsl_y = gsl_vector_view_array(est_tau, *n);

		/*do the fit*/
		gsl_vector *reg_coef = gsl_vector_alloc(ncoef);
		gsl_multifit_linear_workspace *mw = gsl_multifit_linear_alloc(*n, ncoef);
		double chisq;
		gsl_multifit_linear(gsl_X, &gsl_y.vector, reg_coef, gsl_Cov, &chisq, mw);

		/*prediction*/
		double yerr;
		for (int i=0;i<*n;i++){
			gsl_bspline_eval(Ztest[i], gsl_B, gsl_bs);
			gsl_multifit_linear_est(gsl_B, reg_coef, gsl_Cov, &tau_est[i], &yerr);
		}

		/*free the allocation*/
		gsl_vector_free(gsl_bp);
		gsl_vector_free(gsl_Z2);
		gsl_vector_free(gsl_B);
		gsl_vector_free(reg_coef);
		gsl_matrix_free(gsl_X);
		gsl_matrix_free(gsl_Cov);
		gsl_bspline_free(gsl_bs);
		gsl_multifit_linear_free(mw);
	}
	catch (...) {
		::Rf_error( "c++ exception (unknown reason)" );
	}
}

// Tune the number of knots for fitting B-spline
// number of knots up to 5
void CV_BS(double *tau_est, double *Ztest, double *Z, double *est_tau, int *Itrain, int *Itest, int *m, int *n, int *K)
{
	try{
		gsl_vector *gsl_SSR2 = gsl_vector_calloc(6);
		double *SSR2_ptr = gsl_vector_ptr(gsl_SSR2, 0);

		int m0 = *n-*m;

		/*Polynomial fit*/
		int ncoef = 4;
		gsl_multifit_linear_workspace *mw = gsl_multifit_linear_alloc(m0, ncoef);
		
		/*cunstruct the fit matrix and the response, do the fit*/
		gsl_matrix *gsl_X = gsl_matrix_alloc(m0, ncoef);
		gsl_matrix *gsl_Cov = gsl_matrix_alloc(ncoef, ncoef);
		gsl_vector *gsl_B = gsl_vector_alloc(ncoef);
		gsl_vector *gsl_y = gsl_vector_alloc(m0);
		gsl_vector *reg_coef = gsl_vector_alloc(ncoef);
		double yerr;
		double chisq;
		for (int k=0;k<*K;k++){
			for (int i=0;i<m0;i++){
				/*construct the fit matrix*/
				gsl_matrix_set(gsl_X, i, 0, 1.0);
				gsl_matrix_set(gsl_X, i, 1, Z[Itrain[k*(m0)+i]]);
				gsl_matrix_set(gsl_X, i, 2, Z[Itrain[k*(m0)+i]]*Z[Itrain[k*(m0)+i]]);
				gsl_matrix_set(gsl_X, i, 3, Z[Itrain[k*(m0)+i]]*Z[Itrain[k*(m0)+i]]*Z[Itrain[k*(m0)+i]]);
				/*construct the response*/
				gsl_vector_set(gsl_y, i, est_tau[Itrain[k*(m0)+i]]);
			}

			/*do the fit*/
			gsl_multifit_linear(gsl_X, gsl_y, reg_coef, gsl_Cov, &chisq, mw);

			/*prediction*/
			for (int i=0;i<*m;i++){
				gsl_vector_set(gsl_B, 0, 1.0);
				gsl_vector_set(gsl_B, 1, Z[Itest[k*(*m)+i]]);
				gsl_vector_set(gsl_B, 2, Z[Itest[k*(*m)+i]]*Z[Itest[k*(*m)+i]]);
				gsl_vector_set(gsl_B, 3, Z[Itest[k*(*m)+i]]*Z[Itest[k*(*m)+i]]*Z[Itest[k*(*m)+i]]);
				gsl_multifit_linear_est(gsl_B, reg_coef, gsl_Cov, &tau_est[Itest[k*(*m)+i]], &yerr);
				SSR2_ptr[0] = SSR2_ptr[0]+(tau_est[Itest[k*(*m)+i]]-est_tau[Itest[k*(*m)+i]])*(tau_est[Itest[k*(*m)+i]]-est_tau[Itest[k*(*m)+i]]);
			}
		}

		/*spline fit, with number of knots 1,2,3,4,5*/
		for (int nnod=1;nnod<6;nnod++){
			for (int k=0;k<*K;k++){
				BS(tau_est, Z, est_tau, &Itrain[k*(m0)], &Itest[k*(*m)], &nnod, m, n);
				for (int i=0;i<*m;i++){
					SSR2_ptr[nnod] = SSR2_ptr[nnod]+(tau_est[Itest[k*(*m)+i]]-est_tau[Itest[k*(*m)+i]])*(tau_est[Itest[k*(*m)+i]]-est_tau[Itest[k*(*m)+i]]);
				}
			}
		}

		/*return the index with the smallest SSR2*/
		int MI = gsl_vector_min_index(gsl_SSR2);
		if (MI==0){
			gsl_X = gsl_matrix_alloc(*n, ncoef);
			gsl_y = gsl_vector_alloc(*n);
			mw = gsl_multifit_linear_alloc(*n, ncoef);

			for (int i=0;i<*n;i++){
				/*construct the fit matrix*/
				gsl_matrix_set(gsl_X, i, 0, 1.0);
				gsl_matrix_set(gsl_X, i, 1, Z[i]);
				gsl_matrix_set(gsl_X, i, 2, Z[i]*Z[i]);
				gsl_matrix_set(gsl_X, i, 3, Z[i]*Z[i]*Z[i]);
				/*construct the response*/
				gsl_vector_set(gsl_y, i, est_tau[i]);
			}

			/*do the fit*/
			gsl_multifit_linear(gsl_X, gsl_y, reg_coef, gsl_Cov, &chisq, mw);

			/*prediction*/
			for (int i=0;i<*n;i++){
				gsl_vector_set(gsl_B, 0, 1.0);
				gsl_vector_set(gsl_B, 1, Ztest[i]);
				gsl_vector_set(gsl_B, 2, Ztest[i]*Ztest[i]);
				gsl_vector_set(gsl_B, 3, Ztest[i]*Ztest[i]*Ztest[i]);
				gsl_multifit_linear_est(gsl_B, reg_coef, gsl_Cov, &tau_est[i], &yerr);
			}
		}
		else{
			BSII(tau_est, Ztest, Z, est_tau, &MI, n);
		}

		/*free the allocation*/
		gsl_vector_free(gsl_SSR2);
		gsl_vector_free(gsl_y);
		gsl_vector_free(reg_coef);
		gsl_matrix_free(gsl_X);
		gsl_vector_free(gsl_B);
		gsl_matrix_free(gsl_Cov);
	}
	catch (...) {
		::Rf_error( "c++ exception (unknown reason)" );
	}
}

// generate the data dependent sketching matrix
// ITrain, the index of training samples, K*(n/(K-1)*K)
// ITest, the index of testing samples
// no: an integer vector of length 4, 
// no[1]=n, no[2]=K, no[3]=B, no[4]=m, no[5]=p, no[6]=nnod
// no[7] number of observations in testing samples
// no[8]=L 
// p dimensionality of X (n * p)
// n length of A
// K number of cross-validated samples
// B number of generated sketching matrices to take maximum
// m number of generated sketching matrices to take average
// nnod number of nods used in spline basis
// Xtest: the testing matrices of dimension n * p
// est_tau: (A/pi-(1-A)/(1-pi))Y+augmented term
// tau_est_test: estimated contrast function for the testing samples n * m
// p0: the probability of a binomial distribution
void GS(double *tau_est_test, double *Xtest, double *X, double *A, double *Y, double *est_pih, double *est_tau, int *ITrain, int *ITest, int *no, double *p0, int *Iter)
{
	try{
		// n, K, B, m and s2
		int n = no[0];
		int K = no[1];
		int B = no[2];
		int m = no[3];
		int p = no[4];
		int ntest = no[6];
		int m0 = n-ntest;
		int Int[100];
		int Int0[100];
		size_t MI = 0;
		double s2=0;
		
		// Initialize the index set
		for (int i=0;i<p;i++)
			Int[i] = i;
		
		// random number generator
 		const gsl_rng_type *T;
		gsl_rng *r; 

		/*create a generator chosen by the
		environment variable GSL_RNG_TYPE*/
 		gsl_rng_env_setup();
		gsl_rng_default_seed = 123456*(*Iter);
		T = gsl_rng_default;
		r = gsl_rng_alloc (T); 

		// sketching matrices S of dimension p*B, Z of length n and VD.est of length B
		gsl_matrix *gsl_S = gsl_matrix_alloc(B, p);
		gsl_vector *gsl_s = gsl_vector_alloc(p);
		gsl_vector *gsl_Z = gsl_vector_alloc(n); 
		gsl_vector *gsl_Ztest = gsl_vector_alloc(n);
		gsl_vector *gsl_VD = gsl_vector_alloc(B);
		gsl_vector *gsl_tau_est = gsl_vector_alloc(n);
		gsl_vector_view gsl_sub_S;
		gsl_matrix_view gsl_X = gsl_matrix_view_array(X, n, p);
		gsl_matrix_view gsl_Xtest = gsl_matrix_view_array(Xtest, n, p);
		double *S_ptr = gsl_matrix_ptr(gsl_S, 0, 0);
		double *Z_ptr = gsl_vector_ptr(gsl_Z, 0);
		double *Ztest_ptr = gsl_vector_ptr(gsl_Ztest, 0);
		double *VD_ptr = gsl_vector_ptr(gsl_VD, 0);
		double *tau_est_ptr = gsl_vector_ptr(gsl_tau_est, 0);

		// main loop
		for (int i=0;i<m;i++){
			// initialize the sketching matrix
			for (int j=0;j<B;j++){
				for (int k=0;k<p;k++){
					S_ptr[p*j+k] = 0;
				}
			}
			// initialize the value difference
			for (int j=0;j<B;j++){
				VD_ptr[j] = 0;
			}

			for (int j=0;j<B;j++){
				// select a random L
				int L = gsl_ran_binomial(r, *p0, p-2);
				L = L+2;
				
				// select a random sample
				gsl_ran_choose(r, Int0, L, Int, p, sizeof (int));
				
				// generate independent sketching matrices
				s2 = 0;
				for (int k=0;k<L;k++){
					S_ptr[p*j+Int0[k]] = gsl_ran_ugaussian(r);
					s2 = s2+S_ptr[p*j+Int0[k]]*S_ptr[p*j+Int0[k]];
				}
 				// standardize
				for (int k=0;k<L;k++){
					S_ptr[p*j+Int0[k]] = S_ptr[p*j+Int0[k]]/sqrt(s2);
				} 
				
				// obtain the projected covariates
				gsl_sub_S = gsl_matrix_row(gsl_S, j);
				gsl_blas_dgemv(CblasNoTrans, 1.0, &gsl_X.matrix, &gsl_sub_S.vector, 0.0, gsl_Z);

				// cross-validated updates
				for (int k=0;k<K;k++){
					BS(tau_est_ptr, Z_ptr, est_tau, &ITrain[k*m0], &ITest[k*ntest], &no[5], &ntest, &n);
				}

//				// calculate the value difference
				for (int k=0;k<K;k++){
					VD(&VD_ptr[j], A, Y, est_pih, tau_est_ptr, &ITest[k*ntest], &ntest, &n);
				}
//				VD_ptr[j] = VD_ptr[j]-1.0*L*log(1.0*n);
				// calculate the deviance
//				for (int k=0;k<K;k++){
//					for (int l=0;l<ntest;l++){
//						VD_ptr[j] = VD_ptr[j]-(tau_est_ptr[ITest[k*(ntest)+l]]-est_tau[ITest[k*(ntest)+l]])*
//							(tau_est_ptr[ITest[k*(ntest)+l]]-est_tau[ITest[k*(ntest)+l]]);
//						VD_ptr[j] = VD_ptr[j]-ghf(tau_est_ptr[ITest[k*(ntest)+l]], tau_est_ptr[ITest[k*(ntest)+l]]);
//					}
//				}
			}

			// Take the random projection that yields the largest VD
			MI = gsl_vector_max_index(gsl_VD);
			for (int j=0;j<p;j++){
				gsl_vector_set(gsl_s, j, S_ptr[p*(MI)+j]);
			}
			
//			// comment
//			for (int j=0;j<p;j++){
//				S0[j] = VD_ptr[j];
//				S0[p*i+j] = S_ptr[p*(MI)+j];
//			}
			
			// do the projection
			gsl_blas_dgemv(CblasNoTrans, 1.0, &gsl_X.matrix, gsl_s, 0.0, gsl_Z);
			gsl_blas_dgemv(CblasNoTrans, 1.0, &gsl_Xtest.matrix, gsl_s, 0.0, gsl_Ztest);

			// Fitting the B-spline
			CV_BS(tau_est_ptr, Ztest_ptr, Z_ptr, est_tau, ITrain, ITest, &ntest, &n, &K);

			// obtain d_est_test
			for (int j=0;j<n;j++){
				tau_est_test[i*n+j] = tau_est_ptr[j];
			}
		}

		// free the random number generator and sketching matrix
//		gsl_rng_free (r);
		gsl_matrix_free (gsl_S);
		gsl_vector_free (gsl_s);
		gsl_vector_free (gsl_Z);
		gsl_vector_free (gsl_Ztest);
		gsl_vector_free (gsl_VD);
		gsl_vector_free (gsl_tau_est);
	}
	catch (...) {
		::Rf_error( "c++ exception (unknown reason)" );
	}
}

}