## simulation code for testing high dimensional qualitative treatment effects
## Chengchun Shi, April 17, 2017

library(ncvreg)
library(listdtr)
## data generation function
## tau: the specification of contrast function
## if tau=1, tau=2Xbeta I(Xbeta>0)+2min(Xbeta+delta,0)
## else tau=(Xbeta)^2-delta
## return the main subspace
Gen.data <- function(n=500, p=50, tau=1, delta=Inf){

X <- matrix(rnorm(n*p), n, p)

  A <- rbinom(n, 1, 0.5)
  # generate the baseline
  h0 <- 1+(X[,1]-X[,2])/2
  # generate the contrast function
  beta0 <- rep(1, p)/sqrt(p)
  if (tau==1){
    tau0 <- delta*sqrt(pi)*(X[,1]+X[,2])
	tau0 <- tau0*(X[,3]+X[,4]+X[,5]+X[,6]+X[,7])^2/5
  }
  else if (tau==2){
    tau0 <- 0.5*((X[,1]+X[,2])^2-delta)
	tau0 <- tau0*(X[,3]+X[,4]+X[,5]+X[,6]+X[,7])^2/5
  }
  else if (tau==3){
  	tau0 <- delta*(cos((X[,1]+X[,2])*pi/sqrt(2)))
	tau0 <- tau0*(X[,3]+X[,4]+X[,5]+X[,6]+X[,7])^2/5
  }
  else{
  	tau0 <- delta*sqrt(60*pi)*rowMeans(X[,1:30])*(X[,31]+X[,32]+X[,33]+X[,34]+X[,35])^2/5
  }
  
  # generate the response
  Y <- h0+A*tau0+0.5*rnorm(n)
  Z <- X%*%beta0
  
  return(list(Y=Y, X=X, A=A, Z=Z))
}

## estimate the propensity score and baseline function
## using sparse penalized regression with SCAD penalty
## estimate the propensity score and baseline function
## using sparse penalized regression with SCAD penalty
pih <- function(X, A, Y){
  
  p <- dim(X)[2]
  
  # estimate the baseline
  theta <- cv.ncvreg(X[A==0,], Y[A==0], penalty='SCAD', family="gaussian", nfolds=5)
  theta.est <- coef(theta, s='lambda.min')
  theta.est <- theta.est[1:(p+1)]
  h0.est <- cbind(1,X)%*%theta.est
  
  # estimate the baseline
  zeta <- cv.ncvreg(X[A==1,], Y[A==1], penalty='SCAD', family="gaussian", nfolds=5)
  zeta.est <- coef(zeta, s='lambda.min')
  zeta.est <- zeta.est[1:(p+1)]
  h1.est <- cbind(1,X)%*%zeta.est
  
  # estimate the propensity score
  alpha <- cv.ncvreg(X, A, family="binomial", penalty='SCAD', nfolds=5)
  alpha.est <- coef(alpha, s='lambda.min')
  pi.est <- plogis(as.vector(cbind(1,X) %*% as.matrix(alpha.est)))
  
  # construct the residual
  result <- cv.ncvreg(X, Y, penalty='SCAD', family="gaussian", nfolds=5)
  result <- coef(result, s='lambda.min')
  result <- result[1:(p+1)]
  mu.est <- cbind(1,X)%*%theta.est
  
  # obtian residual
  Y <- Y-mu.est
  
  M.alpha <- alpha.est!=0
  M.theta0 <- theta.est!=0
  M.theta1 <- zeta.est!=0
  
  return(list(pi.est=pi.est, h0.est=h0.est, h1.est=h1.est, M.alpha=M.alpha, 
              M.theta0=M.theta0, M.theta1=M.theta1, alpha.est=alpha.est,
              theta.est=theta.est, zeta.est=zeta.est, Y=Y))
}

## construct the pseodu outcome
pseoduY <- function(X, A, Y, est.pih){
  pi.est <- est.pih$pi.est
  h0.est <- est.pih$h0.est
  h1.est <- est.pih$h1.est
  tau.est <- ((1-A)/(1-pi.est)-A/(pi.est))*Y
  tau.est <- tau.est-((1-A)/(1-pi.est)-1)*h0.est+(A/pi.est-1)*h1.est
  
  return(-tau.est)
}

## compute the test statistic
stat <- function(X, A, Y, est.pih, dn, tau.est){
  n <- length(A)
  pi.est <- est.pih$pi.est
  h0.est <- est.pih$h0.est
  h1.est <- est.pih$h1.est
  M.alpha <- est.pih$M.alpha
  M.theta0 <- est.pih$M.theta0
  M.theta1 <- est.pih$M.theta1
  
  ## compute the numerator
  d.est <- (tau.est>0)
  if (dim(tau.est)[2]>1)
    d.est <- rowMeans(d.est)
  else
    d.est <- as.vector(d.est)
  VD.est <- ((1-A)/(1-pi.est)-A/(pi.est))*Y
  VD.est <- VD.est-((1-A)/(1-pi.est)-1)*h0.est+(A/pi.est-1)*h1.est
  #  kappa <- VD.est
  VD.est <- VD.est*(1-d.est)
  kappa <- VD.est
  
  ## compute the denominator
  Sigma2 <- t(cbind(1,X)[,M.theta0]) %*% ((1-A)*cbind(1,X)[,M.theta0])/n
  Sigma3 <- t(cbind(1,X)[,M.theta1]) %*% (A*cbind(1,X)[,M.theta1])/n
  if (length(which(M.theta0))==1)
    barI2 <- mean(as.vector(((1-A)/(1-pi.est)-1)) * (cbind(1,X)[,M.theta0]) * (1-d.est))
  else
    barI2 <- colMeans(as.vector(((1-A)/(1-pi.est)-1)) * (cbind(1,X)[,M.theta0]) * (1-d.est))
  if (length(which(M.theta1))==1)
    barI3 <- mean(as.vector(A/pi.est-1) * (cbind(1,X)[,M.theta1]) * (1-d.est))
  else
    barI3 <- colMeans(as.vector(A/pi.est-1) * (cbind(1,X)[,M.theta1]) * (1-d.est))
  SiginvbarI2 <- solve(Sigma2, barI2)
  SiginvbarI3 <- solve(Sigma3, barI3)
  
  if (length(SiginvbarI2)>1)
    kappa <- kappa-(as.vector((1-A)*(Y-h0.est))*(cbind(1,X)[,M.theta0]))%*%SiginvbarI2
  else
    kappa <- kappa-(as.vector((1-A)*(Y-h0.est))*(cbind(1,X)[,M.theta0]))*SiginvbarI2
  if (length(SiginvbarI3)>1)
    kappa <- kappa+(as.vector(A*(Y-h1.est))*(cbind(1,X)[,M.theta1]))%*%SiginvbarI3
  else
    kappa <- kappa+(as.vector(A*(Y-h1.est))*(cbind(1,X)[,M.theta1]))*SiginvbarI3
  if (length(which(M.alpha))>0){
    Sigma1 <- t(cbind(1,X)[,M.alpha]) %*% (pi.est*(1-pi.est)*cbind(1,X)[,M.alpha])/n
    if (length(which(M.alpha))==1){
      barI1 <- mean(as.vector((pi.est*(1-A)/(1-pi.est))*(Y-h0.est)+((1-pi.est)*A/pi.est)*(Y-h1.est))
                    *cbind(1,X)[,M.alpha]*(1-d.est))
    }
    else{
      barI1 <- colMeans(as.vector((pi.est*(1-A)/(1-pi.est))*(Y-h0.est)+((1-pi.est)*A/pi.est)*(Y-h1.est))
                        *cbind(1,X)[,M.alpha]*(1-d.est))
    }
    SiginvbarI1 <- solve(Sigma1, barI1)
    if (length(SiginvbarI1)>1)
      kappa <- kappa+(A-pi.est)*cbind(1,X)[,M.alpha]%*%SiginvbarI1
    else
      kappa <- kappa+(A-pi.est)*cbind(1,X)[,M.alpha]*SiginvbarI1
  }
  #  kappa <- kappa*(1-d.est)
  sig.est <- max(sqrt(var(kappa)), dn)
  
  return(sqrt(n)*mean(VD.est)/sig.est)
}

ISS <- function(X, Y, A, step){
  #some constants
  n <- length(Y)
  p <- dim(X)[2]
  
  # V_est: use the Q-learning method to estimate
  V_est<-matrix(0, nrow=n, ncol=step+1)
  a_opt <- matrix(0, nrow=n, ncol=(step+1))
  S <- rep(0,step+1)
  regime_coef <- matrix(0, nrow=(step+1), ncol=(p+1))
  
  reg0 <- lm(Y~A)
  a_opt[,1] <- as.numeric(reg0$coefficients[2]>=0)
  regime_coef[1,] <- c(as.numeric(reg0$coefficients[2]>=0),rep(0,p))
  V_est[,1] <- reg0$coef[1]+reg0$coef[2]*a_opt[,1]
  S[1] <- sum(V_est[,1]-Y)
  
  #select the most important variable
  #S score
  S_temp <- rep(0,p)
  coef_all <- matrix(0,nrow=p,ncol=2)
  V_temp <- matrix(0,nrow=n,ncol=p)

  for(j in 1:p){
    reg <- lm(Y~X[,j]*A)
    coef <- reg$coef
    coef_all[j,] <- coef[3:4]
    diff <- coef[3]+coef[4]*X[,j]
    g_opt <- as.numeric(diff>=0)
    S_temp[j] <- sum(diff*(g_opt-a_opt[,1]))
	V_temp[,j] <- Y+diff*(g_opt-A)
  }

  #the most important variable
  imp <- which.max(S_temp)
  coef_temp <- coef_all[imp,]
  regime_coef[2,c(1,imp+1)] <- coef_temp
  a_opt[,2] <- as.numeric(cbind(1,X[,imp])%*%coef_temp>0)
  V_est[,2] <- V_temp[,imp]
  S[2] <- S_temp[imp]
  
  #select the rest important variables sequentially
  for (j in 2:step){
    S_temp <- rep(0,p) #in the jth search
    coef_all <- matrix(0,nrow=p,ncol=j+1)
	V_temp <- matrix(0,nrow=n,ncol=p)
    for (k in (1:p)[-imp]){
      #search the maximum, a_opt is the current A regime
      X_temp <- X[,c(imp,k)]
      reg <- lm(Y~X[,c(imp,k)]*A)
      coef <- reg$coef
      coef_all[k,] <- reg$coef[(j+2):(2*(j+1))]
      diff <- cbind(1,X[,c(imp,k)])%*%coef_all[k,]
      g_opt <- as.numeric(diff>=0)
      S_temp[k] <- sum(diff*(g_opt-a_opt[,j]))
	  V_temp[,k] <- Y+diff*(g_opt-A)
    }
    #important variables
    j_opt<-which.max(S_temp)
    imp<-c(imp,j_opt)
    coef_temp<-coef_all[j_opt,]
    regime_coef[j+1,c(1,imp+1)]<-coef_temp
    a_opt[,j+1]<-as.numeric(cbind(1,X[,imp])%*%coef_temp>0)
    V_est[,j+1]<-V_temp[,j_opt]
	S[j+1]<-S_temp[j_opt]
  }
  
  sp <- which.min(-log(cumsum(S))+(0:step)*log(n)/n)
  return (as.vector(regime_coef[sp, ])!=0)
}

## Test based on sequential advantage selection
test.SAS <- function(Model, dn){
  N <- length(Model$A)
  n <- N/2
  p <- dim(Model$X)[2]
  
  ## partition the data into two subsets
  I1 <- sample(1:N, n)
  I2 <- setdiff(1:N, I1)
  
  ## baseline and propensity score
  est.pih1 <- pih(Model$X[I1,], Model$A[I1], Model$Y[I1])
  est.pih2 <- pih(Model$X[I2,], Model$A[I2], Model$Y[I2])
  
  ## estimate the contrast function on the first half of data
  re1 <- ISS(Model$X[I1,], Model$Y[I1], Model$A[I1], step=50)
  re2 <- ISS(Model$X[I2,], Model$Y[I2], Model$A[I2], step=50)
  X.tilde1 <- Model$X[I1, re1[-1]]
  X.tilde2 <- Model$X[I2, re2[-1]]
  Y.pseodu1 <- pseoduY(Model$X[I1,], Model$A[I1], Model$Y[I1], est.pih1)
  Y.pseodu2 <- pseoduY(Model$X[I2,], Model$A[I2], Model$Y[I2], est.pih2)
  if (length(which(re1))>1){ 
    obj1 <- krr(X.tilde1, Y.pseodu1)
    tau.est.all <- predict(obj1, Model$X[I2, re1[-1]])
  }
  else{
    tau.est.all <- as.matrix(rep(mean(Y.pseodu2), n))
  }
  
  # construct the test statistic
  stat1 <- stat(Model$X[I2,], Model$A[I2], Model$Y[I2], est.pih2, dn, tau.est.all)
  
  ## estimate the contrast function on the second half of data
  if (length(which(re2))>1){ 
    obj2 <- krr(X.tilde2, Y.pseodu2)
    tau.est.all <- predict(obj2, Model$X[I1, re2[-1]])
  }
  else{
    tau.est.all <- as.matrix(rep(mean(Y.pseodu1), n))
  }
  
  # construct the test statistic
  stat2 <- stat(Model$X[I1,], Model$A[I1], Model$Y[I1], est.pih1, dn, tau.est.all)
  
  # take the supremum of two test statistics
  stat0 <- max(stat1, stat2)
  return(stat0)
}

# GO, PACK!
stat0 <- rep(0, 500)
dn <- log(log(500, 10))/(500^(1/6))
p <- 50
for (l in 1:500) {
  #do your work
  print(l)
  
  #sleep to make this take extra time
  set.seed(12345*l)
  
  Model <- Gen.data(n=500, p=p, tau=2, delta=2*0.3722)
  stat0[l] <- test.SAS(Model, dn)
  
  #save the data every 50 iterations
  if (l %% 50 == 0) {
    #save all of the data
    save.image(file="SAS2.RData")
    
    print("saved")
  }	
}
